#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <dos.h>
#include "qic02.h"

extern int  streamer(int, void*, int);
extern long ptr2abs(void far *);

#define dim(x) (sizeof(x)/sizeof(*(x)))
#define eol(s) (*(s)=='\n'||*(s)=='\0')
#define MAXBLOCK 32
#define TBLOCK   512

static struct init_data id = {0x300, 3, -1};
static WORD block = 1;
static BYTE *buffer = NULL;
static int  buflen = 0;
static WORD bufadr = 0;

static void *dmabuf(int length, int /* size of DMA word */dma_word)
{
   register void *ptr;

   ptr = malloc(length);
   if (!ptr) {
      (void)printf("Not enough memory\n");
   } else if (dma_word > 0) {
      long a; register s;

      a = ptr2abs(ptr);
      s = dma_word + 15; /* shift factor to get DMA page number */
      /* Compare DMA pages */
      if ((a >> s) != ((a+length) >> s)) {
      /* Attempt to get DMA page-aligned buffer assumes: */
      /*  - length is not greater than half of DMA page; */
      /*  - malloc() heap is unfragmented.               */
         register void *tmp;

         tmp = ptr; ptr = malloc(length); free(tmp);
         if (!ptr) {
            (void)printf("No memory to align buffer\n");
            return (void *)0;
         }
         a = ptr2abs(ptr);
         if ((a >> s) != ((a+length) >> s)) {
            (void)printf("Fail to align buffer\n");
            free(ptr); return (void *)0;
         }
      }
      if (a & (dma_word - 1)) {
         (void)printf("DMA buffer is not word-aligned\n");
         free(ptr); return (void *)0;
      }
   }
   return ptr;
}

static void prerror(int k)
{
   if (!k) {
      printf("Done.     \n");
   } else if (k == CTE_FAULT) {
      perror("Software error");
   } else if (k < qic02_nerr) {
      printf("Error: %s\n", qic02_errlist[k]);
   } else {
      printf("Error N%d\n", k);
   }
}

static void test(char *s)
{
   register k;
   register unsigned m;
   struct qic02_word_status q;

   if (!eol(s)) {
      (void)printf("Extra parameters\n"); return;
   }
   if ((k = streamer('t', &q, 0)) >= CTE_BUSY) {
      prerror(k);
   } else {
      printf("Got status: %04xh\n", q.status);
      for (k=16; k-- > 0;) {
         if (q.status & (m = 1 << k)) {
            printf("%04x: %s\n", m, qic02_errlist[k+1]);
         }
      }
   }
}

static void cmdrew(char *s)
{
   if (!eol(s)) {
      (void)printf("Extra parameters\n"); return;
   }
   (void)printf("Working...\r");
   fflush(stdout);
   prerror(streamer('n', NULL, 0));
}

static void retens(char *s)
{
   if (!eol(s)) {
      (void)printf("Extra parameters\n"); return;
   }
   (void)printf("Working...\r");
   fflush(stdout);
   prerror(streamer('i', NULL, Q2_RETENSION));
}

static void ctread(char *s)
{
   register k;

   if (!eol(s)) {
      (void)printf("Extra parameter(s)\n"); return;
   }
   bufadr = 0;
   buflen = TBLOCK*block;
   (void)printf("Working...\r");
   fflush(stdout);

   k = streamer('r', buffer, buflen);
   if (k != buflen) {
      prerror(streamer('t', NULL, 0));
   } else {
      printf("\n");
   }
   if (k >= 0) {
      (void)printf("%d (%xh) bytes read\n", k, k);
      if (k < buflen) buflen = k;
   }
}

static void settings(void)
{
   printf(
      "Configuration set: base = %3xh   dma = %o   irq = %-2d   block = %d\n",
      id.base_address, id.dma_number, id.irq_number, block);
}

static int setint(WORD *dest, char *s, int inf, int sup)
{
   register i;

   if (eol(s)) return 0;
   for (i=0; *s >= '0' && *s<='9'; s++) {
       if ((i = 10*i + *s - '0') > sup) goto bad;
   }
   if (eol(s) && i>=inf) {
      *dest = i; return 0;
   }
bad:
   printf("Invalid parameter: must be %d-%d\n", inf, sup);
   return -1;
}

static int sethex(WORD *dest, char *s, int inf, int sup)
{
   register i, k;

   if (eol(s)) return 0;
   i = 0;
   for (i=0; ;s++) {
      if      (*s >= '0' && *s<='9') k = '0';
      else if (*s >= 'a' && *s<='f') k = 'a' - 10;
      else if (*s >= 'A' && *s<='F') k = 'A' - 10;
      else {
         if (*s == 'h' || *s=='H') s += 1;
         break;
      }
      if ((i = (i<<4) | (*s-k)) > sup) goto bad;
   }
   if (eol(s) && i>=inf) {
      *dest = i; return 0;
   }
bad:
   printf("Invalid parameter: must be %xh-%xh\n", inf, sup);
   return -1;
}

static int setval(WORD *dest, char *s, int inf, int sup)
{
   register i, k;

   for (i=0, k=0; !eol(s+i) && !k; i++) {
      if (s[i]>='a' && s[i]<='f' || s[i] == 'h' ||
          s[i]>='A' && s[i]<='F' || s[i] == 'H') k = 1;
   }
   return k ? sethex(dest, s, inf, sup) : setint(dest, s, inf, sup);
}

static void skip(char *s)
{
   register i;
   WORD n = 1;

   if (setval(&n, s, 1, 0x4000) != 0) return;

   printf("Working...\r");
   fflush(stdout);

   if ((i = streamer('f', NULL, n)) != n) {
      prerror(streamer('t', NULL, 0));
      printf("%d file(s) skipped\n", i);
   } else {
      prerror(0);
   }
}

static void dump(char *s)
{
   int a, i, j;

   if (buflen == 0) {
      (void)printf("Buffer is empty\n"); return;
   }
   if (bufadr >= buflen) bufadr = 0;
   if (setval(&bufadr, s, 0, buflen-1) != 0) return;

   a = bufadr & ~15;
   for (j=0; j<8 && a<buflen; j++) {
      printf("%04xh", a);
      for (i=0; i<16 && a+i<buflen; i++) {
         if (a+i >= bufadr) {
            printf(i==7 ? "-" : " ");
            printf("%02.2x", buffer[a+i]);
         } else {
            printf("   ");
         }
      }
      printf("   ");
      for (i=0; i<16 && a+i<buflen; i++) {
         if (a+i >= bufadr) {
            printf("%c", buffer[a+i] & ~037 ? buffer[a+i] : '.');
         } else {
            printf(" ");
         }
      }
      printf("\n");
      bufadr = (a += i);
   }
}

void setblk(char *s)
{
   register WORD old = block;
   if (setint(&block, s, 1, MAXBLOCK) == 0) settings();
   if (block != old) {
      buflen = 0;
      if (buffer) free(buffer);
      buffer = dmabuf(TBLOCK*block, id.dma_number > 3 ? 2 : 1);
      if (!buffer) exit(-1);
   }
}

void setbas(char *s)
{
   register k;
   register old = id.base_address;

   if (sethex(&(id.base_address), s, 0, 0x3ff) == 0) settings();
   if (id.base_address != old) {
      (void)streamer('c', NULL, 0);
      if ((k = streamer('o', &id, 0)) != 0) prerror(k);
   }
}

void setirq(char *s)
{
   register k;
   register old = id.irq_number;

   if (setint(&(id.irq_number), s, 0, 15) == 0) settings();
   if (id.irq_number != old) {
      (void)streamer('c', NULL, 0);
      if ((k = streamer('o', &id, 0)) != 0) prerror(k);
   }
}

void setdma(char *s)
{
   register k;
   register old = id.dma_number;

   if (setint(&(id.dma_number), s, 0, 7) == 0) settings();
   if (id.dma_number != old) {
      (void)streamer('c', NULL, 0);
      if ((k = streamer('o', &id, 0)) != 0) prerror(k);
   }
}

static void finish(char *s)
{
   (void)s; (void)streamer('c', NULL, 0); exit(0);
}

main(int argc, char *argv[])
{
   register i, j, k, l;
   register char *p;
   static struct { char *name; void (*func)(char *); } cmdlist[] = {
      { "block",     setblk },
      { "base",      setbas },
      { "dump",      dump   },
      { "dma",       setdma },
      { "get",       ctread },
      { "irq",       setirq },
      { "quit",      finish },
      { "qic",       NULL   },
      { "reset",     NULL   },
      { "rewind",    cmdrew },
      { "retension", retens },
      { "skip",      skip   },
      { "test",      test   },
   };
   char line[81];

   if (argc != 2) {
      printf("Usage: streamer <device_name>\n\n\
Device names are:\n\teverex, fastape, wangtek\n");
      exit(-1);
   }
   if ((j = streamer('?', argv[1], 0)) < 0) {
      printf("Invalid device name\n");
      exit(-1);
   }
/* id.dma_number = 1; */
   buffer = dmabuf(TBLOCK*block, id.dma_number > 3 ? 2 : 1);
   if (!buffer) exit(-1);
   settings();

   if ((k = streamer('o', &id, j)) != 0) prerror(k);

   for (line[sizeof(line)-1]='\0'; ;) {
      printf("=> ");
      fgets(line, sizeof(line)-1, stdin);
      for (l=0; l<sizeof(line) && (line[l] == ' ' || line[l] == '\t'); l++);
      for (i=l; i<sizeof(line) &&
                line[i] != ' ' && line[i] != '\n' && line[i] != '\t'; i++);
      if (i == l) goto next;
      for (k=i; k<sizeof(line) && (line[k] == ' ' || line[k] == '\t'); k++);

      for (p=line+l, j=0; j<dim(cmdlist); j++) {
         if (strnicmp(p, cmdlist[j].name, i) == 0) {
            if (cmdlist[j].func) {
               cmdlist[j].func(line+k);
            } else {
               printf("Sorry, \'%s\' not implemented yet\n", cmdlist[j].name);
            }
            goto next;
         }
      }
      printf ("Bad command \'%-*.*s\'\n", i-l, i-l, p);
next:;
   }
}
